from __future__ import annotations

import csv
import enum
import json
import sys
from typing import Any, Dict, Iterable, Iterator, List, Optional, Sequence, cast

from rich import box
from rich.console import Console
from rich.live import Live
from rich.table import Table


class OutputFormat(enum.Enum):
    table = "table"
    csv = "csv"


def as_table(dicts: Sequence[Dict[str, Any]], keys: Optional[List[str]] = None) -> List[List[Optional[str]]]:
    """
    Converts a list of dictionaries to a list of lists (table), ordered by specified keys.

    :param dicts: list of dictionaries
    :param keys: ordered list of keys to include in each row, or None to use the keys from the first dict
    :return: list of lists, with None values if there's no value for a key
    """
    if not dicts:
        return []

    if keys is None:
        keys = list(dicts[0].keys())
    return [keys] + [[str(d.get(f, "")) if d.get(f, "") else None for f in keys] for d in dicts]  # type: ignore


def as_strings(values: Iterable[Any]) -> List[str]:
    return [str(v) if v else "" for v in values]


def pretty_print(
    result: List[Dict[str, Any]] | Iterator[Dict[str, Any]] | Dict | str | None,
    output_format: OutputFormat = OutputFormat.table,
) -> None:
    """print results as table/csv/json."""

    console = Console()

    if isinstance(result, list) and not result:
        console.print("No results")
        return

    elif isinstance(result, list) and output_format == OutputFormat.table:
        rows = as_table(result)
        column_names = cast(List[str], rows[0])
        table = Table(box=box.SIMPLE)
        for c in column_names:
            if c in ["CommandId"]:
                table.add_column(c, no_wrap=True)
            else:
                table.add_column(c)

        for r in rows[1:]:
            table.add_row(*r)

        console.print(table)

    elif isinstance(result, list) and output_format == OutputFormat.csv:
        writer = csv.DictWriter(sys.stdout, fieldnames=list(result[0].keys()))

        writer.writeheader()
        for r in result:
            writer.writerow(r)

    elif isinstance(result, Iterator) and output_format == OutputFormat.table:
        try:
            first = next(result)
        except StopIteration:
            console.print("No results")
            return

        table = Table(box=box.SIMPLE)
        for c in first.keys():
            table.add_column(c)

        table.add_row(*as_strings(first.values()))

        with Live(table, refresh_per_second=1):
            for row in result:
                table.add_row(*as_strings(row.values()))

    elif isinstance(result, Iterator) and output_format == OutputFormat.csv:
        writer = csv.writer(sys.stdout)
        first = next(result)
        writer.writerow(first.keys())
        writer.writerow(first.values())
        for row in result:
            writer.writerow(row.values())

    elif isinstance(result, dict):
        print(json.dumps(result, default=str))

    elif not result:
        print("Done ✨")

    else:
        print(result)



# HandledErrors are caught and their message printed without a stack trace
from typing import Optional


class HandledError(Exception):
    pass


class NoInstancesError(HandledError):
    def __init__(self, name: Optional[str] = None, name_match: Optional[str] = None):
        if name:
            pretty_name = "instance id {name}" if name.startswith("i-") else f"name {name}"
            msg = f"No instances with {pretty_name}"
        elif name_match:
            msg = f"No instances with name matching {name_match}"
        else:
            raise ValueError("Missing name or name_match")

        super(NoInstancesError, self).__init__(msg)



import os.path
from argparse import Namespace
from typing import Any, Callable, Dict, List, Optional, Union

import pytoml as toml
from typing_extensions import TypedDict


class SsmConfig(TypedDict, total=False):
    s3bucket: str
    s3prefix: str


class VpcConfig(TypedDict, total=False):
    name: str
    subnet: str
    security_group: str


class Config(TypedDict, total=False):
    ssm: SsmConfig
    vpc: VpcConfig
    key_name: str
    # the following fields are optional
    region: str
    additional_tags: Dict[str, str]
    iam_instance_profile_arn: str
    kms_key_id: str
    describe_images_owners: Union[List[str], str]
    describe_images_name_match: str
    launch_template: str
    volume_size: int


def inject_config(config_file: str) -> Callable[[Namespace], None]:
    """Replace the "config" arg value with a dict loaded from the config file."""

    def inner(namespace: Namespace) -> None:
        # replace the "config" arg value with a dict loaded from the config file
        if "config" in namespace:
            setattr(namespace, "config", load_config(config_file, namespace.config))

    return inner


# TODO add tests for this
def load_config(config_file: str, profile_override: Optional[str] = None) -> Config:
    """
    Load profile from the config file.

    :param config_file: path to config file
    :param profile_override: override the value of the default profile in the config file
    :raises Exception: if problem loading the config
    :return: config dictionary
    """
    config_filepath = os.path.expanduser(config_file)
    config = load_user_config_file(config_filepath)

    profile = profile_override
    if not profile:
        # set prof to the value of the default key
        if "default_profile" not in config:
            raise Exception(f"No profile override supplied, or default profile set in {config_filepath}")
        profile = config["default_profile"]

    # make top level keys available in the profile
    if config.get("additional_tags", None):
        config[profile]["additional_tags"] = config["additional_tags"]

    try:
        return config[profile]
    except KeyError:
        raise Exception(f"Missing profile {profile_override} in {config_filepath}")


def load_user_config_file(config_filepath: str) -> Dict[str, Any]:
    if not os.path.isfile(config_filepath):
        raise Exception(f"No config file {config_filepath}")

    with open(config_filepath) as config_file:
        return toml.load(config_file)



import argparse
import sys
import traceback
from typing import List

from botocore.exceptions import ClientError, NoCredentialsError

import aec.command.ami as ami
import aec.command.compute_optimizer as compute_optimizer
import aec.command.ec2 as ec2
import aec.command.ssm as ssm
import aec.util.cli as cli
import aec.util.config as config
import aec.util.configure as configure
import aec.util.display as display
from aec.util.cli import Arg, Cmd, parameter_defaults
from aec.util.errors import HandledError

config_arg = Arg("--config", help="Section of the config file to use")


def ami_arg_checker(s: str) -> str:
    if not (s.startswith("ami-") or s in ami.ami_keywords.keys()):
        raise argparse.ArgumentTypeError(f"must begin with 'ami-' or be one of {list(ami.ami_keywords.keys())}")

    return s


def tag_arg_checker(tag: str) -> str:
    parts = tag.split("=")
    if len(parts) != 2:
        raise argparse.ArgumentTypeError(f"Invalid tag argument '{tag}'. Must be in key=value form.")

    return tag


# fmt: off

configure_cli = [
    Cmd(configure.example)
]

ec2_cli = [
    Cmd(ec2.create_key_pair, [
        config_arg,
        Arg("key_name", type=str, help="Key pair name"),
        Arg("file_path", type=str,help="Path of private key to create, eg: ~/.ssh/topsecret.pem"),
    ]),
    Cmd(ec2.describe, [
        config_arg,
        Arg("name", type=str, nargs='?', help="Filter to instances with this Name tag or instance id."),
        Arg("-q", type=str, dest='name_match', help="Filter to instances with a Name tag containing NAME_MATCH."),
        Arg("-r", "--show-running-only", action='store_true', help="Show running or pending instances only"),
        Arg("-it", "--include-terminated", action='store_true', help="Include terminated instances"),
        Arg("-s", "--sort-by", type=str, help="Sort by one or more fields", default=parameter_defaults(ec2.describe)["sort_by"]),
        Arg("-c", "--columns", type=str, help="Customise the columns shown", default=parameter_defaults(ec2.describe)["columns"]),
    ]),
    Cmd(ec2.launch, [
        config_arg,
        Arg("name", type=str, help="Name tag of instance"),
        Arg("-a", "--ami", type=ami_arg_checker, help=f"AMI id or a keyword to lookup the latest ami: {[k for k in ami.ami_keywords.keys()]}"),
        Arg("-t", "--template", type=str, help="Launch template name"),
        Arg("--volume-size", type=int, help="EBS volume size (GB). Defaults to AMI volume size."),
        Arg("--encrypted", type=bool, help="Whether the EBS volume is encrypted", default=True),
        Arg("--instance-type", type=str, help="Instance type"),
        Arg("--key-name", type=str, help="Key name"),
        Arg("--userdata", type=str, help="User data file path or http URL")
    ]),
    Cmd(ec2.logs, [
        config_arg,
        Arg("name", type=str, help="Name tag of instance or instance id")
    ]),
    Cmd(ec2.modify, [
        config_arg,
        Arg("name", type=str, help="Name tag of instance or instance id"),
        Arg("type", type=str, help="Type of instance")
    ]),
    Cmd(ec2.start, [
        config_arg,
        Arg("name", type=str, help="Name tag of instance or instance id")
    ]),
    Cmd(ec2.stop, [
        config_arg,
        Arg("name", type=str, help="Name tag of instance or instance id")
    ]),
    Cmd(ec2.tag, [
        config_arg,
        Arg("-t", "--tags", type=tag_arg_checker, nargs='+', metavar="TAG", help="Tags to create in key=value form", required = True),
        Arg("name", type=str, nargs='?', help="Filter to instances with this Name tag or instance id."),
        Arg("-q", type=str, dest='name_match', help="Filter to instances with a Name tag containing NAME_MATCH.")
    ]),
    Cmd(ec2.describe_tags, [
        config_arg,
        Arg("name", type=str, nargs='?', help="Filter to instances with this Name tag or instance id."),
        Arg("-q", type=str, dest='name_match', help="Filter to instances with a Name tag containing NAME_MATCH."),
        Arg("-v", "--volumes", action='store_true', help="Show volumes"),
        Arg("-k", "--keys", type=str, nargs='*', metavar="KEY", help="Tags to display", default = []),
    ], name = "tags"),
    Cmd(ec2.status, [
        config_arg
    ]),
    Cmd(ec2.templates, [
        config_arg
    ]),
    Cmd(ec2.terminate, [
        config_arg,
        Arg("name", type=str, help="Name tag of instance or instance id")
    ]),
    Cmd(ec2.user_data, [
        config_arg,
        Arg("name", type=str, help="Name tag of instance or instance id"),
    ]),
]

ami_cli = [
    Cmd(ami.delete, [
        config_arg,
        Arg("ami", type=str, help="AMI id")
    ]),
    Cmd(ami.describe, [
        config_arg,
        Arg("id", type=str, nargs='?', help="Filter to this AMI id"),
        Arg("--owner", type=str, help="Filter to this owning account"),
        Arg("-q", type=str, dest='name_match', help="Filter to images with a name containing NAME_MATCH."),
        Arg("--show-snapshot-id", action='store_true', help="Show snapshot id")
    ]),
    Cmd(ami.share, [
        config_arg,
        Arg("ami", type=str, help="AMI id"),
        Arg("account", type=str, help="Account id"),
    ])
]

compute_optimizer_cli = [
    Cmd(compute_optimizer.over_provisioned, [
        config_arg
    ])
]

ssm_cli = [
    Cmd(ssm.commands, [
        config_arg,
        Arg("name", type=str, nargs='?', help="Filter to instances with this Name tag or instance id."),
    ]),
    Cmd(ssm.compliance_summary, [
        config_arg
    ]),
    Cmd(ssm.describe, [
        config_arg,
        Arg("name", type=str, nargs='?', help="Filter to instances with this Name tag or instance id."),
        Arg("-q", type=str, dest='name_match', help="Filter to instances with a Name tag containing NAME_MATCH."),
    ]),
    Cmd(ssm.invocations, [
        config_arg,
        Arg("command_id", type=str, help="Command id"),
    ]),
    Cmd(ssm.output, [
        config_arg,
        Arg("command_id", type=str, help="Command id"),
        Arg("instance_id", type=str, help="Instance id"),
        Arg("-e", "--stderr", action='store_true', help="Show stderr instead of stdout"),
    ]),
    Cmd(ssm.patch, [
        config_arg,
        Arg("operation", type=str, choices=["scan", "install"], help="Scan or install"),
        Arg("names", type=str, nargs='+', help="Name tag of instance or instance id. Use 'all' for all running instances"),
        Arg("-nr","--no-reboot", action='store_true', help="Do not reboot after install"),
    ]),
    Cmd(ssm.patch_summary, [
        config_arg
    ]),
    Cmd(ssm.run, [
        config_arg,
        Arg("names", type=str, nargs='+', help="Name tags of instance or instance ids. Use 'all' for all running instances.")
    ]),
]
# fmt: on


def build_parser() -> argparse.ArgumentParser:
    # prog=aec so that we run in cog the program name isn't cog
    parser = argparse.ArgumentParser(prog="aec", description="aws ec2 cli")
    subparsers = parser.add_subparsers(title="commands")

    cli.add_command_group(subparsers, "configure", "Configure subcommands", configure_cli)
    cli.add_command_group(subparsers, "ec2", "EC2 subcommands", ec2_cli, config.inject_config("~/.aec/ec2.toml"))
    cli.add_command_group(subparsers, "ami", "AMI subcommands", ami_cli, config.inject_config("~/.aec/ec2.toml"))
    cli.add_command_group(
        subparsers,
        "co",
        "Compute optimizer subcommands",
        compute_optimizer_cli,
        config.inject_config("~/.aec/ec2.toml"),
    )
    cli.add_command_group(subparsers, "ssm", "SSM subcommands", ssm_cli, config.inject_config("~/.aec/ec2.toml"))

    return parser


def main(args: List[str] = sys.argv[1:]) -> None:
    try:
        result, output_format = cli.dispatch(build_parser(), args)
        display.pretty_print(result, output_format)
    except ClientError as e:
        code = e.response["Error"]["Code"]
        if code == "UnauthorizedOperation":
            message = e.response["Error"]["Message"]
            print(f"{code}: {message}\n\nAuthenticate with the appropriate AWS role before retrying.", file=sys.stderr)

        elif code == "RequestExpired":
            print(
                f"{code}: AWS session token expired.\n\nRe-authenticate with the appropriate AWS role.", file=sys.stderr
            )

        else:
            traceback.print_exc(file=sys.stderr)
    except NoCredentialsError as e:
        print(
            f"NoCredentialsError: {e}.\n\nAuthenticate with the appropriate AWS role before retrying.", file=sys.stderr
        )
    except HandledError as e:
        print(e, file=sys.stderr)


if __name__ == "__main__":
    main()



from __future__ import annotations

import base64
import os
import os.path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, cast

import boto3
from typing_extensions import TypedDict

from aec.util.ec2 import describe_running_instances_names
from aec.util.errors import NoInstancesError
from aec.util.threads import executor

if TYPE_CHECKING:
    from mypy_boto3_ec2.type_defs import (
        BlockDeviceMappingTypeDef,
        FilterTypeDef,
        TagTypeDef,
        InstanceStatusSummaryTypeDef,
        TagSpecificationTypeDef,
    )
    from mypy_boto3_ec2.literals import InstanceTypeType

import aec.command.ami as ami_cmd
import aec.util.ec2 as util_tags
from aec.util.config import Config
from aec.util.ec2_types import RunArgs


def is_ebs_optimizable(instance_type: str) -> bool:
    return not instance_type.startswith("t2")


class Instance(TypedDict, total=False):
    InstanceId: str
    State: str
    Name: Optional[str]
    Type: str
    DnsName: str
    SubnetId: str


def launch(
    config: Config,
    name: str,
    ami: Optional[str] = None,
    template: Optional[str] = None,
    volume_size: Optional[int] = None,
    encrypted: bool = True,
    instance_type: Optional[str] = None,
    key_name: Optional[str] = None,
    userdata: Optional[str] = None,
) -> List[Instance]:
    """Launch a tagged EC2 instance with an EBS volume."""

    template = template or config.get("launch_template", None)

    if not (template or ami):
        raise ValueError("Please specify either an ami or a launch template")

    if not template and not instance_type:
        # if no instance type is provided set one
        instance_type = "t3.small"

    ec2_client = boto3.client("ec2", region_name=config.get("region", None))

    runargs: RunArgs = {
        "MaxCount": 1,
        "MinCount": 1,
    }

    desc = ""
    if template:
        runargs["LaunchTemplate"] = {"LaunchTemplateName": template}
        desc = f"template {template} "

    if ami:
        image = ami_cmd.fetch(config, ami)
        if not image["RootDeviceName"]:
            raise ValueError(f"{image['ImageId']} is missing RootDeviceName")

        volume_size = volume_size or config.get("volume_size", None) or image["Size"]
        if not volume_size:
            raise ValueError("No volume size")

        device: BlockDeviceMappingTypeDef = {
            "DeviceName": image["RootDeviceName"],
            "Ebs": {
                "VolumeSize": volume_size,
                "DeleteOnTermination": True,
                "VolumeType": "gp3",
                "Encrypted": encrypted,
            },
        }

        kms_key_id = config.get("kms_key_id", None)
        if kms_key_id:
            device["Ebs"]["KmsKeyId"] = kms_key_id

        runargs["ImageId"] = image["ImageId"]
        runargs["BlockDeviceMappings"] = [device]

        desc = desc + image["Name"] if image["Name"] else desc + image["ImageId"]

    key = key_name or config.get("key_name", None)
    if key:
        runargs["KeyName"] = key
    elif not template:
        print(
            "WARNING: You have not specified a key pair.",
            "You will only be able to connect to this instance if it is configured for EC2 Instance Connect or Systems Manager Session Manager.",
        )

    if instance_type:
        runargs["InstanceType"] = cast("InstanceTypeType", instance_type)
        runargs["EbsOptimized"] = is_ebs_optimizable(instance_type)

    tags: List[TagTypeDef] = [{"Key": "Name", "Value": name}]
    additional_tags = config.get("additional_tags", {})
    if additional_tags:
        tags.extend([{"Key": k, "Value": v} for k, v in additional_tags.items()])
    runargs["TagSpecifications"] = [
        cast("TagSpecificationTypeDef", {"ResourceType": "instance", "Tags": tags}),
        cast("TagSpecificationTypeDef", {"ResourceType": "volume", "Tags": tags}),
    ]

    if config.get("vpc", None):
        # TODO: support multiple subnets
        security_group = config["vpc"]["security_group"]
        runargs["NetworkInterfaces"] = [
            {
                "DeviceIndex": 0,
                "Description": "Primary network interface",
                "DeleteOnTermination": True,
                "SubnetId": config["vpc"]["subnet"],
                "Ipv6AddressCount": 0,
                "Groups": security_group if isinstance(security_group, list) else [security_group],
            }
        ]
        associate_public_ip_address = config["vpc"].get("associate_public_ip_address", None)
        if associate_public_ip_address is not None:
            runargs["NetworkInterfaces"][0]["AssociatePublicIpAddress"] = associate_public_ip_address

        vpc_name = f" vpc {config['vpc']['name']}"
    else:
        vpc_name = ""

    iam_instance_profile_arn = config.get("iam_instance_profile_arn", None)
    if iam_instance_profile_arn:
        runargs["IamInstanceProfile"] = {"Arn": iam_instance_profile_arn}

    if userdata:
        if userdata.startswith("http://") or userdata.startswith("https://"):
            runargs["UserData"] = read_url(userdata)
        else:
            runargs["UserData"] = read_file(userdata)

    # use IMDSv2 to prevent SSRF
    runargs["MetadataOptions"] = {"HttpTokens": "required"}

    region_name = ec2_client.meta.region_name

    print(f"Launching a {instance_type} in {region_name}{vpc_name} named {name} using {desc} ... ")
    response = ec2_client.run_instances(**runargs)

    instance = response["Instances"][0]

    waiter = ec2_client.get_waiter("instance_running")
    waiter.wait(InstanceIds=[instance["InstanceId"]])

    # TODO: wait until instance checks passed (as they do in the console)

    # the response from run_instances above always contains an empty string
    # for PublicDnsName, so we call describe to get it
    return describe(config=config, name=name)


def describe(
    config: Config,
    name: Optional[str] = None,
    name_match: Optional[str] = None,
    include_terminated: bool = False,
    show_running_only: bool = False,
    sort_by: str = "State,Name",
    columns: str = "InstanceId,State,Name,Type,DnsName,LaunchTime,ImageId",
) -> List[Instance]:
    """List EC2 instances in the region."""

    ec2_client = boto3.client("ec2", region_name=config.get("region", None))

    filters = name_filters(name, name_match)
    if show_running_only:
        filters.append({"Name": "instance-state-name", "Values": ["pending", "running"]})

    # TODO: handle more than 1000 results by looping using NextToken
    response = ec2_client.describe_instances(MaxResults=1000, Filters=filters)

    # print(response["Reservations"][0]["Instances"][0])

    cols = columns.split(",")

    # don't sort by cols we aren't showing
    sort_cols = [sc for sc in sort_by.split(",") if sc in cols]

    instances: List[Instance] = []
    for r in response["Reservations"]:
        for i in r["Instances"]:
            if include_terminated or i["State"]["Name"] != "terminated":
                desc: Instance = {}

                for col in cols:
                    if col == "State":
                        desc[col] = i["State"]["Name"]
                    elif col == "Name":
                        desc[col] = util_tags.get_value(i, "Name")
                    elif col == "Type":
                        desc[col] = i["InstanceType"]
                    elif col == "DnsName":
                        desc[col] = i["PublicDnsName"] if i.get("PublicDnsName", None) != "" else i["PrivateDnsName"]
                    else:
                        desc[col] = i.get(col, None)

                instances.append(desc)

    return sorted(
        instances,
        key=lambda i: "".join(str(i[field]) for field in sort_cols),
    )


def describe_tags(
    config: Config,
    name: Optional[str] = None,
    name_match: Optional[str] = None,
    keys: List[str] = [],
    volumes: bool = False,
) -> List[Dict[str, Any]]:
    """List EC2 instances or volumes with their tags."""
    if volumes:
        return volume_tags(config, name, name_match, keys)

    return instance_tags(config, name, name_match, keys)


def tag(
    config: Config,
    tags: List[str],
    name: Optional[str] = None,
    name_match: Optional[str] = None,
) -> List[Dict[str, Any]]:
    """Tag EC2 instance(s)."""
    ec2_client = boto3.client("ec2", region_name=config.get("region", None))

    tagdefs: List[TagTypeDef] = []
    for t in tags:
        parts = t.split("=")
        tagdefs.append({"Key": parts[0], "Value": parts[1]})

    if not name and not name_match:
        # avoid tagging all instances when there's no name
        raise NoInstancesError(name=name, name_match=name_match)

    instances = describe(config, name, name_match)

    ids = [i["InstanceId"] for i in instances]

    if not ids:
        raise NoInstancesError(name=name, name_match=name_match)

    ec2_client.create_tags(Resources=ids, Tags=tagdefs)

    return describe_tags(config, name, name_match, keys=[d["Key"] for d in tagdefs])


def instance_tags(
    config: Config, name: Optional[str] = None, name_match: Optional[str] = None, keys: List[str] = []
) -> List[Dict[str, Any]]:
    """List EC2 instances with their tags."""

    ec2_client = boto3.client("ec2", region_name=config.get("region", None))

    response = ec2_client.describe_instances(Filters=name_filters(name, name_match))

    instances: List[Dict[str, Any]] = []
    for r in response["Reservations"]:
        for i in r["Instances"]:
            if i["State"]["Name"] != "terminated":
                inst = {"InstanceId": i["InstanceId"], "Name": util_tags.get_value(i, "Name")}
                if not keys:
                    inst["Tags"] = ", ".join(f"{tag['Key']}={tag['Value']}" for tag in i.get("Tags", []))
                else:
                    for key in keys:
                        inst[f"Tag: {key}"] = util_tags.get_value(i, key)

                instances.append(inst)

    return sorted(instances, key=lambda i: str(i["Name"]))


def volume_tags(
    config: Config, name: Optional[str] = None, name_match: Optional[str] = None, keys: List[str] = []
) -> List[Dict[str, Any]]:
    """List EC2 volumes with their tags."""

    ec2_client = boto3.client("ec2", region_name=config.get("region", None))

    response = ec2_client.describe_volumes(Filters=name_filters(name, name_match))

    volumes: List[Dict[str, Any]] = []
    for v in response["Volumes"]:
        vol = {"VolumeId": v["VolumeId"], "Name": util_tags.get_value(v, "Name")}
        if not keys:
            vol["Tags"] = ", ".join(f"{tag['Key']}={tag['Value']}" for tag in v.get("Tags", []))
        else:
            for key in keys:
                vol[f"Tag: {key}"] = util_tags.get_value(v, key)

        volumes.append(vol)

    return sorted(volumes, key=lambda i: str(i["Name"]))


def start(config: Config, name: str) -> List[Instance]:
    """Start EC2 instance."""

    ec2_client = boto3.client("ec2", region_name=config.get("region", None))

    print(f"Starting instances with the name {name} ... ")

    if not name:
        # avoid starting all instances when there's no name
        raise NoInstancesError(name=name)

    instances = describe(config, name)

    if not instances:
        raise NoInstancesError(name=name)

    instance_ids = [instance["InstanceId"] for instance in instances]
    ec2_client.start_instances(InstanceIds=instance_ids)

    waiter = ec2_client.get_waiter("instance_running")
    waiter.wait(InstanceIds=instance_ids)

    return describe(config, name)


def stop(config: Config, name: str) -> List[Dict[str, Any]]:
    """Stop EC2 instance."""

    ec2_client = boto3.client("ec2", region_name=config.get("region", None))

    if not name:
        # avoid stopping all instances when there's no name
        raise NoInstancesError(name=name)

    instances = describe(config, name)

    if not instances:
        raise NoInstancesError(name=name)

    response = ec2_client.stop_instances(InstanceIds=[instance["InstanceId"] for instance in instances])

    return [{"State": i["CurrentState"]["Name"], "InstanceId": i["InstanceId"]} for i in response["StoppingInstances"]]


def terminate(config: Config, name: str) -> List[Dict[str, Any]]:
    """Terminate EC2 instance."""

    ec2_client = boto3.client("ec2", region_name=config.get("region", None))

    if not name:
        # avoid terminating all instances when there's no name
        raise NoInstancesError(name=name)

    instances = describe(config, name)

    if not instances:
        raise NoInstancesError(name=name)

    response = ec2_client.terminate_instances(InstanceIds=[instance["InstanceId"] for instance in instances])

    return [
        {"State": i["CurrentState"]["Name"], "InstanceId": i["InstanceId"]} for i in response["TerminatingInstances"]
    ]


def modify(config: Config, name: str, type: str) -> List[Instance]:
    """Change an instance's type."""

    ec2_client = boto3.client("ec2", region_name=config.get("region", None))

    if not name:
        # avoid modifying all instances when there's no name
        raise NoInstancesError(name=name)

    instances = describe(config, name)

    if not instances:
        raise NoInstancesError(name=name)

    instance_id = instances[0]["InstanceId"]
    ec2_client.modify_instance_attribute(InstanceId=instance_id, InstanceType={"Value": type})
    ec2_client.modify_instance_attribute(InstanceId=instance_id, EbsOptimized={"Value": is_ebs_optimizable(type)})

    return describe(config, name)


def create_key_pair(config: Config, key_name: str, file_path: str) -> str:
    """Create a key pair."""

    ec2_client = boto3.client("ec2", region_name=config.get("region", None))

    path = os.path.expanduser(file_path)
    with open(path, "x") as file:
        key = ec2_client.create_key_pair(KeyName=key_name)
        file.write(key["KeyMaterial"])
        os.chmod(path, 0o600)

    return f"Created key pair {key_name} and saved private key to {path}"


def logs(config: Config, name: str) -> str:
    """Show the system logs."""

    ec2_client = boto3.client("ec2", region_name=config.get("region", None))

    if not name:
        # avoid describing all instances when there's no name
        raise NoInstancesError(name=name)

    instances = describe(config, name)

    if not instances:
        raise NoInstancesError(name=name)

    instance_id = instances[0]["InstanceId"]
    response = ec2_client.get_console_output(InstanceId=instance_id)

    return response.get("Output", "No logs yet 😔")


def templates(config: Config) -> List[Dict[str, Any]]:
    """Describe launch templates."""

    ec2_client = boto3.client("ec2", region_name=config.get("region", None))

    response = ec2_client.describe_launch_templates()

    return [
        {"Name": t["LaunchTemplateName"], "Default Version": t["DefaultVersionNumber"]}
        for t in response["LaunchTemplates"]
    ]


def status(config: Config) -> List[Dict[str, Any]]:
    """Describe instances status checks."""
    ec2_client = boto3.client("ec2", region_name=config.get("region", None))

    kwargs: Dict[str, Any] = {"MaxResults": 50}

    response_fut = executor.submit(ec2_client.describe_instance_status, **kwargs)
    instances = executor.submit(describe_running_instances_names, config).result()
    response = response_fut.result()

    statuses = []
    while True:
        statuses.extend(
            [
                {
                    "InstanceId": i["InstanceId"],
                    "State": i["InstanceState"]["Name"],
                    "Name": instances.get(i["InstanceId"], None),
                    "System status check": status_text(i["SystemStatus"]),
                    "Instance status check": status_text(i["InstanceStatus"]),
                }
                for i in response["InstanceStatuses"]
            ]
        )

        next_token = response.get("NextToken", None)
        if next_token:
            kwargs["NextToken"] = next_token
            response = ec2_client.describe_instance_status(**kwargs)
        else:
            break

    return sorted(
        statuses,
        key=lambda i: "".join(str(i[field]) for field in ["State", "Name"]),
    )


def status_text(summary: InstanceStatusSummaryTypeDef, key: str = "reachability") -> str:
    status = [d for d in summary["Details"] if d["Name"] == key][0]
    return f"{status['Name']} {status['Status']}" + (
        f" since {status['ImpairedSince']}" if status.get("ImpairedSince", None) else ""
    )


def user_data(config: Config, name: str) -> Optional[str]:
    """Describe user data for an instance."""
    ec2_client = boto3.client("ec2", region_name=config.get("region", None))

    if not name:
        # avoid describing all instances when there's no name
        raise NoInstancesError(name=name)

    instances = describe(config, name)

    if not instances:
        raise NoInstancesError(name=name)

    instance_id = instances[0]["InstanceId"]
    response = ec2_client.describe_instance_attribute(Attribute="userData", InstanceId=instance_id)

    try:
        return base64.b64decode(response["UserData"]["Value"]).decode("utf-8")
    except KeyError:
        return None


def name_filters(name_or_id: Optional[str] = None, name_match: Optional[str] = None) -> List[FilterTypeDef]:
    if name_or_id and name_or_id.startswith("i-"):
        return [{"Name": "instance-id", "Values": [name_or_id]}]
    elif name_or_id:
        return [{"Name": "tag:Name", "Values": [name_or_id]}]
    elif name_match:
        return [{"Name": "tag:Name", "Values": [f"*{name_match}*"]}]
    else:
        return []


def read_file(filepath: str) -> str:
    with open(os.path.expanduser(filepath)) as file:
        return file.read()


def read_url(url: str) -> str:
    import requests

    r = requests.get(url)
    r.raise_for_status()
    return r.text



from __future__ import annotations

from typing import TYPE_CHECKING, List, Sequence

if TYPE_CHECKING:
    from mypy_boto3_ec2.literals import InstanceTypeType, ShutdownBehaviorType
    from mypy_boto3_ec2.type_defs import (
        BlockDeviceMappingTypeDef,
        CapacityReservationSpecificationTypeDef,
        CpuOptionsRequestTypeDef,
        CreditSpecificationRequestTypeDef,
        ElasticGpuSpecificationTypeDef,
        ElasticInferenceAcceleratorTypeDef,
        EnclaveOptionsRequestTypeDef,
        FilterTypeDef,
        HibernationOptionsRequestTypeDef,
        IamInstanceProfileSpecificationTypeDef,
        InstanceIpv6AddressTypeDef,
        InstanceMarketOptionsRequestTypeDef,
        InstanceMetadataOptionsRequestTypeDef,
        InstanceNetworkInterfaceSpecificationTypeDef,
        LaunchTemplateSpecificationTypeDef,
        LicenseConfigurationRequestTypeDef,
        PlacementTypeDef,
        RunInstancesMonitoringEnabledTypeDef,
        TagSpecificationTypeDef,
    )

from typing_extensions import TypedDict


class _RequiredRunArgs(TypedDict, total=True):
    MaxCount: int
    MinCount: int


class RunArgs(_RequiredRunArgs, total=False):
    BlockDeviceMappings: List["BlockDeviceMappingTypeDef"]
    ImageId: str
    InstanceType: InstanceTypeType
    Ipv6AddressCount: int
    Ipv6Addresses: List["InstanceIpv6AddressTypeDef"]
    KernelId: str
    KeyName: str
    Monitoring: "RunInstancesMonitoringEnabledTypeDef"
    Placement: "PlacementTypeDef"
    RamdiskId: str
    SecurityGroupIds: List[str]
    SecurityGroups: List[str]
    SubnetId: str
    UserData: str
    AdditionalInfo: str
    ClientToken: str
    DisableApiTermination: bool
    DryRun: bool
    EbsOptimized: bool
    IamInstanceProfile: "IamInstanceProfileSpecificationTypeDef"
    InstanceInitiatedShutdownBehavior: ShutdownBehaviorType
    NetworkInterfaces: List["InstanceNetworkInterfaceSpecificationTypeDef"]
    PrivateIpAddress: str
    ElasticGpuSpecification: List["ElasticGpuSpecificationTypeDef"]
    ElasticInferenceAccelerators: List[ElasticInferenceAcceleratorTypeDef]
    TagSpecifications: List["TagSpecificationTypeDef"]
    LaunchTemplate: LaunchTemplateSpecificationTypeDef
    InstanceMarketOptions: InstanceMarketOptionsRequestTypeDef
    CreditSpecification: "CreditSpecificationRequestTypeDef"
    CpuOptions: CpuOptionsRequestTypeDef
    CapacityReservationSpecification: CapacityReservationSpecificationTypeDef
    HibernationOptions: HibernationOptionsRequestTypeDef
    LicenseSpecifications: List[LicenseConfigurationRequestTypeDef]
    MetadataOptions: InstanceMetadataOptionsRequestTypeDef
    EnclaveOptions: EnclaveOptionsRequestTypeDef


class DescribeArgs(TypedDict, total=False):
    Filters: Sequence["FilterTypeDef"]
    InstanceIds: Sequence[str]
    DryRun: bool
    MaxResults: int
    NextToken: str



"""Helper functions for describing and building a CLI with command groups, which contain many subcommands."""

import inspect
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser, Namespace, _SubParsersAction
from typing import Any, Callable, Dict, List, Optional, Tuple

from aec.util.display import OutputFormat


class Arg:
    def __init__(self, *args: Any, **kwargs: Any):
        self.args = args
        self.kwargs = kwargs


class Cmd:
    def __init__(
        self,
        call_me: Callable[..., Any],
        args: Optional[List[Arg]] = None,
        name: Optional[str] = None,
        help: Optional[str] = None,
    ):
        # dispatch() will call this function
        self.call_me = call_me
        # args for ArgumentParser
        self.args = args
        # fallback to call_me's name in kebab case
        self.name: str = self.call_me.__name__.replace("_", "-") if name is None else name
        # fallback to call_me's docstring for help
        self.help = inspect.getdoc(self.call_me) if help is None else help

        # check all function params have args specified
        call_me_num_args = len(inspect.signature(self.call_me).parameters)
        if self.args and len(self.args) != call_me_num_args:
            raise Exception(
                f"{self.call_me.__name__} has {call_me_num_args} args but {len(self.args)} defined for the cli"
            )
        elif not self.args and call_me_num_args > 0:
            raise Exception(f"{self.call_me.__name__} has {call_me_num_args} args but none defined for the cli")


def usage_exit(parser: ArgumentParser) -> Callable[[], None]:
    def inner() -> None:
        parser.print_usage()
        parser.exit(1, "{}: no subcommand specified\n".format(parser.prog))

    return inner


def add_command_group(
    parent: _SubParsersAction,
    name: str,
    help: str,
    cmds: List[Cmd],
    args_pre_processor: Optional[Callable[[Namespace], None]] = None,
) -> None:
    group = parent.add_parser(name, help=help)
    # show help if no args provided to the command group
    group.set_defaults(call_me=usage_exit(group))
    if args_pre_processor:
        group.set_defaults(args_pre_processor=args_pre_processor)
    subcommands = group.add_subparsers(title="subcommands")

    for cmd in cmds:
        parser = subcommands.add_parser(
            cmd.name, help=cmd.help, description=cmd.help, formatter_class=ArgumentDefaultsHelpFormatter
        )
        parser.set_defaults(call_me=cmd.call_me)
        if cmd.args:
            for arg in cmd.args:
                parser.add_argument(*arg.args, **arg.kwargs)

        # add output arg to every command
        parser.add_argument(
            "-o", "--output", choices=OutputFormat.__members__, help="Output format", default=OutputFormat.table.value
        )


def dispatch(parser: ArgumentParser, args: List[str]) -> Tuple[Any, OutputFormat]:
    pargs = parser.parse_args(args)

    if "args_pre_processor" in pargs:
        pargs.args_pre_processor(pargs)
        delattr(pargs, "args_pre_processor")

    if "call_me" not in pargs:
        parser.print_usage()
        parser.exit(1, "{}: no command specified\n".format(parser.prog))

    call_me = pargs.call_me
    # remove call_me arg because the call_me function doesn't expect it
    delattr(pargs, "call_me")

    # remove output because that's injected above and the call_me function doesn't expect it
    if "output" not in pargs:
        # no subcommand specified
        output_format = OutputFormat.table
    else:
        output_format = OutputFormat[pargs.output]
        delattr(pargs, "output")

    return (call_me(**vars(pargs)), output_format)


def parameter_defaults(func: Callable) -> Dict[str, Any]:
    """
    Get a function's parameter defaults.

    Args:
        func (Callable): function

    Returns:
        Dict[str, Any]: Dictionary of parameter name => default value or None
    """
    signature = inspect.signature(func)
    return {k: v.default for k, v in signature.parameters.items() if v.default is not inspect.Parameter.empty}



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Any, Dict, List

import boto3
import pytz
from dateutil import relativedelta

if TYPE_CHECKING:
    from mypy_boto3_compute_optimizer.type_defs import UtilizationMetricTypeDef

from aec.util.config import Config


def over_provisioned(config: Config) -> List[Dict[str, Any]]:
    """Show recommendations for over-provisioned EC2 instances."""

    def util(metric: UtilizationMetricTypeDef) -> str:
        return f'{metric["name"]} {metric["statistic"][:3]} {metric["value"]}'

    instances_uptime = describe_instances_uptime(config)

    client = boto3.client("compute-optimizer", region_name=config.get("region", None))

    response = client.get_ec2_instance_recommendations(filters=[{"name": "Finding", "values": ["Overprovisioned"]}])

    recs = [
        {
            "ID": i["instanceArn"].split("/")[1],
            "Name": i.get("instanceName", None),
            "Instance Type": i["currentInstanceType"],
            "Recommendation": i["recommendationOptions"][0]["instanceType"],
            "Utilization": util(i["utilizationMetrics"][0]),
            "Uptime": instances_uptime[i["instanceArn"].split("/")[1]],
        }
        for i in response["instanceRecommendations"]
    ]

    return recs


def describe_instances_uptime(config: Config) -> Dict[str, str]:
    """List EC2 instance uptimes in the region."""

    ec2_client = boto3.client("ec2", region_name=config.get("region", None))

    response = ec2_client.describe_instances()

    instances = {
        i["InstanceId"]: difference_in_words(datetime.now(pytz.utc), i["LaunchTime"])
        for r in response["Reservations"]
        for i in r["Instances"]
    }

    return instances


def difference_in_words(date1: datetime, date2: datetime) -> str:
    difference = relativedelta.relativedelta(date1, date2)

    words = ""

    if difference.months > 0:
        words += f"{difference.months} months "
    if difference.days > 0:
        words += f"{difference.days} days "
    if difference.hours > 0:
        words += f"{difference.hours} hours "
    if words == "":
        if difference.minutes != 0:
            words = f"{difference.minutes} minutes"
        else:
            words = f"{difference.seconds} seconds"

    return words



from __future__ import annotations

from typing import TYPE_CHECKING, Dict, Optional, Sequence

import boto3

from aec.util.config import Config
from aec.util.ec2_types import DescribeArgs

if TYPE_CHECKING:
    from mypy_boto3_ec2.type_defs import InstanceTypeDef, VolumeTypeDef


def get_value(instance: InstanceTypeDef | VolumeTypeDef, key: str) -> Optional[str]:
    tag_value = [t["Value"] for t in instance.get("Tags", []) if t["Key"] == key]
    return tag_value[0] if tag_value else None


def describe_running_instances_names(config: Config) -> Dict[str, Optional[str]]:
    # 2x speed up (8 -> 4 secs) compared to listing all names
    return describe_instances_names(config, {"instance-state-name": ["running"]})


def describe_instances_names(
    config: Config, filters: Optional[Dict[str, Sequence[str]]] = None
) -> Dict[str, Optional[str]]:
    """Map of EC2 instance ids to names in the region."""
    ec2_client = boto3.client("ec2", region_name=config.get("region", None))

    kwargs: DescribeArgs = {"MaxResults": 1000}
    if filters:
        kwargs["Filters"] = [{"Name": k, "Values": v} for k, v in filters.items()]

    instances: Dict[str, Optional[str]] = {}

    while True:
        response = ec2_client.describe_instances(**kwargs)

        for r in response["Reservations"]:
            for i in r["Instances"]:
                instance_id = i["InstanceId"]
                instances[instance_id] = get_value(i, "Name")

        next_token = response.get("NextToken", None)
        if next_token:
            kwargs["NextToken"] = next_token
        else:
            break

    return instances



from __future__ import annotations

from typing import TYPE_CHECKING, List, NamedTuple, Optional

import boto3

if TYPE_CHECKING:
    from mypy_boto3_ec2.type_defs import FilterTypeDef

from typing_extensions import NotRequired, TypedDict

from aec.util.config import Config


class Image(TypedDict):
    Name: Optional[str]
    ImageId: str
    CreationDate: str
    RootDeviceName: Optional[str]
    Size: Optional[int]
    SnapshotId: NotRequired[str]


class AmiMatcher(NamedTuple):
    owner: str
    match_string: str


amazon_base_account_id = "137112412989"
canonical_account_id = "099720109477"

ami_keywords = {
    "amazon2": AmiMatcher(amazon_base_account_id, "amzn2-ami-hvm*x86_64-gp2"),
    "ubuntu1604": AmiMatcher(canonical_account_id, "ubuntu/images/hvm-ssd/ubuntu-xenial-16.04-amd64"),
    "ubuntu1804": AmiMatcher(canonical_account_id, "ubuntu/images/hvm-ssd/ubuntu-bionic-18.04-amd64"),
    "ubuntu2004": AmiMatcher(canonical_account_id, "ubuntu/images/hvm-ssd/ubuntu-focal-20.04-amd64"),
}


def fetch(config: Config, ami: str) -> Image:
    ami_matcher = ami_keywords.get(ami, None)
    if ami_matcher:
        try:
            # lookup the latest ami by name match
            ami_details = describe(config, owner=ami_matcher.owner, name_match=ami_matcher.match_string)[0]
        except IndexError:
            raise RuntimeError(
                f"Could not find ami with name matching {ami_matcher.match_string} owned by account {ami_matcher.owner}"
            )
    else:
        try:
            # lookup by ami id
            ami_details = describe(config, id=ami)[0]
        except IndexError:
            raise RuntimeError(f"Could not find {ami}")
    return ami_details


def describe(
    config: Config,
    id: Optional[str] = None,
    owner: Optional[str] = None,
    name_match: Optional[str] = None,
    show_snapshot_id: bool = False,
) -> List[Image]:
    """List AMIs."""

    ec2_client = boto3.client("ec2", region_name=config.get("region", None))

    if id:
        response = ec2_client.describe_images(ImageIds=[id])
    else:
        if owner:
            owners_filter = [owner]
        else:
            describe_images_owners = config.get("describe_images_owners", None)

            if not describe_images_owners:
                owners_filter = ["self"]
            elif isinstance(describe_images_owners, str):
                owners_filter = [describe_images_owners]
            else:
                owners_filter: List[str] = describe_images_owners

        if name_match is None:
            name_match = config.get("describe_images_name_match", None)

        filters: List[FilterTypeDef] = [] if name_match is None else [{"Name": "name", "Values": [f"*{name_match}*"]}]

        print(f'Describing images owned by {owners_filter} with name matching {name_match or "*"}')

        response = ec2_client.describe_images(Owners=owners_filter, Filters=filters)

    images = []
    for i in response["Images"]:
        image: Image = {
            "Name": i.get("Name", None),
            "ImageId": i["ImageId"],
            "CreationDate": i["CreationDate"],
            "RootDeviceName": i["RootDeviceName"] if "RootDeviceName" in i else None,
            "Size": i["BlockDeviceMappings"][0]["Ebs"]["VolumeSize"] if i["BlockDeviceMappings"] else None,
        }
        if show_snapshot_id:
            image["SnapshotId"] = i["BlockDeviceMappings"][0]["Ebs"]["SnapshotId"]
        images.append(image)

    return sorted(images, key=lambda i: i["CreationDate"], reverse=True)


def delete(config: Config, ami: str) -> None:
    """Deregister an AMI and delete its snapshot."""

    ec2_client = boto3.client("ec2", region_name=config.get("region", None))

    response = describe(config, ami, show_snapshot_id=True)

    ec2_client.deregister_image(ImageId=ami)

    ec2_client.delete_snapshot(SnapshotId=response[0]["SnapshotId"])


def share(config: Config, ami: str, account: str) -> None:
    """Share an AMI with another account."""

    ec2_client = boto3.client("ec2", region_name=config.get("region", None))

    ec2_client.modify_image_attribute(
        ImageId=ami,
        LaunchPermission={"Add": [{"UserId": account}]},
        OperationType="add",
        UserIds=[account],
        Value="string",
        DryRun=False,
    )



from __future__ import annotations

import codecs
import sys
import uuid
from typing import IO, TYPE_CHECKING, Any, Dict, Iterator, List, Optional, Sequence, TypeVar, Union, cast

import boto3
from botocore.exceptions import ClientError
from typing_extensions import Literal, TypedDict

from aec.util.config import Config
from aec.util.ec2 import describe_instances_names, describe_running_instances_names

if TYPE_CHECKING:
    from mypy_boto3_ssm.type_defs import InstanceInformationStringFilterTypeDef


class Agent(TypedDict):
    ID: str
    Name: Optional[str]
    PingStatus: str
    Platform: str
    AgentVersion: Optional[str]


def describe(
    config: Config,
    name: Optional[str] = None,
    name_match: Optional[str] = None,
) -> Iterator[Agent]:
    """List running instances with the SSM agent."""

    instances_names = describe_running_instances_names(config)

    if name:
        filters = name_filters([name])
    elif name_match:
        # unlike ec2 describe_instances, ssm describe_instance_information doesn't
        # support a wildcard name filter. So do the name match here.
        names = [n for n in instances_names.values() if n and name_match in n]
        if not names:
            return
        filters = name_filters(names)
    else:
        filters = []

    kwargs: Dict[str, Any] = {"MaxResults": 50, "Filters": filters}
    client = boto3.client("ssm", region_name=config.get("region", None))
    while True:
        response = client.describe_instance_information(**kwargs)

        for i in response["InstanceInformationList"]:
            a: Agent = {
                "ID": i["InstanceId"],
                "Name": instances_names.get(i["InstanceId"], None),
                "PingStatus": i["PingStatus"],
                "Platform": f'{i["PlatformName"]} {i["PlatformVersion"]}',
                "AgentVersion": i["AgentVersion"],
            }
            yield a

        next_token = response.get("NextToken", None)
        if next_token:
            kwargs["NextToken"] = next_token
        else:
            break


def patch_summary(config: Config) -> Iterator[Dict[str, Any]]:
    """Patch summary for all instances that have run the patch baseline."""
    instances_names = describe_instances_names(config)
    instance_ids = list(instances_names.keys())

    client = boto3.client("ssm", region_name=config.get("region", None))

    max_at_a_time = 50

    for i in range(0, len(instance_ids), max_at_a_time):
        chunk = instance_ids[i : i + max_at_a_time]
        response = client.describe_instance_patch_states(InstanceIds=chunk)
        for i in response["InstancePatchStates"]:
            yield {
                "InstanceId": i["InstanceId"],
                "Name": instances_names.get(i["InstanceId"], None),
                "Needed": i["MissingCount"],
                "Pending Reboot": i["InstalledPendingRebootCount"],
                "Errored": i["FailedCount"],
                "Rejected": i["InstalledRejectedCount"],
                "Last operation time": i["OperationEndTime"],
                "Last operation": i["Operation"],
            }


def compliance_summary(config: Config) -> List[Dict[str, Any]]:
    """Compliance summary for running instances that have run the patch baseline."""
    instances_names = describe_instances_names(config)

    client = boto3.client("ssm", region_name=config.get("region", None))

    response = client.list_resource_compliance_summaries(
        Filters=[{"Key": "ComplianceType", "Values": ["Patch"], "Type": "EQUAL"}]
    )

    return [
        {
            "InstanceId": i["ResourceId"],
            "Name": instances_names.get(i["ResourceId"], None),
            "Status": i["Status"],
            "NonCompliantCount": i["NonCompliantSummary"]["NonCompliantCount"],
            "Last operation time": i["ExecutionSummary"]["ExecutionTime"],
        }
        for i in response["ResourceComplianceSummaryItems"]
    ]


def patch(
    config: Config, operation: Literal["scan", "install"], names: List[str], no_reboot: bool
) -> List[Dict[str, Optional[str]]]:
    """Scan or install AWS patch baseline."""

    instance_ids = fetch_instance_ids(config, names)

    client = boto3.client("ssm", region_name=config.get("region", None))

    kwargs: Dict[str, Any] = {
        "DocumentName": "AWS-RunPatchBaseline",
        "InstanceIds": instance_ids,
    }

    if operation == "scan":
        kwargs["Parameters"] = {"Operation": ["Scan"], "SnapshotId": [str(uuid.uuid4())]}
    else:
        kwargs["Parameters"] = {
            "Operation": ["Install"],
            "RebootOption": ["NoReboot" if no_reboot else "RebootIfNeeded"],
            "SnapshotId": [str(uuid.uuid4())],
        }

    try:
        kwargs["OutputS3BucketName"] = config["ssm"]["s3bucket"]
        kwargs["OutputS3KeyPrefix"] = config["ssm"]["s3prefix"]
    except KeyError:
        pass

    response = client.send_command(**kwargs)

    return [
        {
            "CommandId": response["Command"]["CommandId"],
            "InstanceId": i,
            "Status": response["Command"]["Status"],
            "Document": response["Command"]["DocumentName"],
            "Output": f"s3://{response['Command']['OutputS3BucketName']}/{response['Command']['OutputS3KeyPrefix']}"
            if response["Command"].get("OutputS3BucketName", None)
            else None,
        }
        for i in response["Command"]["InstanceIds"]
    ]


def run(config: Config, names: List[str]) -> List[Dict[str, Optional[str]]]:
    """
    Run a shell script on instance(s).

    Script is read from stdin.
    """

    instance_ids = fetch_instance_ids(config, names)

    client = boto3.client("ssm", region_name=config.get("region", None))

    script = sys.stdin.readlines()

    kwargs: Dict[str, Any] = {
        "DocumentName": "AWS-RunShellScript",
        "InstanceIds": instance_ids,
        "Parameters": {"commands": script},
    }

    try:
        kwargs["OutputS3BucketName"] = config["ssm"]["s3bucket"]
        kwargs["OutputS3KeyPrefix"] = config["ssm"]["s3prefix"]
    except KeyError:
        pass

    response = client.send_command(**kwargs)

    return [
        {
            "CommandId": response["Command"]["CommandId"],
            "InstanceId": i,
            "Status": response["Command"]["Status"],
            "Document": response["Command"]["DocumentName"],
            "Output": f"s3://{response['Command']['OutputS3BucketName']}/{response['Command']['OutputS3KeyPrefix']}"
            if response["Command"].get("OutputS3BucketName", None)
            else None,
        }
        for i in response["Command"]["InstanceIds"]
    ]


E = TypeVar("E")


def first(xs: Optional[Sequence[E]]) -> Optional[E]:
    if xs:
        return xs[0]
    else:
        return None


def commands(config: Config, name: Optional[str] = None) -> Iterator[Dict[str, Union[str, int, None]]]:
    """List commands by instance."""

    client = boto3.client("ssm", region_name=config.get("region", None))

    kwargs: Dict[str, Any] = {"MaxResults": 50}

    if name:
        kwargs["InstanceId"] = fetch_instance_id(config, name)

    while True:
        response = client.list_commands(**kwargs)

        for c in response["Commands"]:
            yield {
                "RequestedDateTime": c["RequestedDateTime"].strftime("%Y-%m-%d %H:%M"),
                "CommandId": c["CommandId"],
                "Status": c["Status"],
                "DocumentName": c["DocumentName"],
                "Operation": first(c["Parameters"].get("Operation", None)),
                "# target": c["TargetCount"],
                "# error": c["ErrorCount"],
                "# timeout": c["DeliveryTimedOutCount"],
                "# complete": c["CompletedCount"],
            }

        next_token = response.get("NextToken", None)
        if next_token:
            kwargs["NextToken"] = next_token
        else:
            break


def invocations(config: Config, command_id: str) -> List[Dict[str, Any]]:
    """List invocations of a command across instances."""

    client = boto3.client("ssm", region_name=config.get("region", None))
    region = client.meta.region_name

    command = client.list_commands(CommandId=command_id)["Commands"][0]
    invocations = client.list_command_invocations(CommandId=command_id)
    instances_names = describe_instances_names(config)

    return [
        {
            "RequestedDateTime": i["RequestedDateTime"].strftime("%Y-%m-%d %H:%M"),
            "InstanceId": i["InstanceId"],
            "Name": instances_names.get(i["InstanceId"], None),
            "StatusDetails": i["StatusDetails"],
            "DocumentName": i["DocumentName"],
            "Operation": ",".join(command["Parameters"]["Operation"]),
            "ConsoleLink": f"https://{region}.console.aws.amazon.com/systems-manager/run-command/{command_id}/{i['InstanceId']}?region={region}",
        }
        for i in invocations["CommandInvocations"]
    ]


DOC_PATHS = {"AWS-RunPatchBaseline": "PatchLinux", "AWS-RunShellScript": "0.awsrunShellScript"}


def output(config: Config, command_id: str, instance_id: str, stderr: bool) -> None:
    """Fetch output of a command from S3."""
    ssm_client = boto3.client("ssm", region_name=config.get("region", None))

    command = ssm_client.list_commands(CommandId=command_id)["Commands"][0]

    if not command.get("OutputS3BucketName", None):
        raise ValueError("No OutputS3BucketName")

    bucket = command["OutputS3BucketName"]

    try:
        doc_path = DOC_PATHS[command["DocumentName"]]
    except KeyError:
        raise NotImplementedError(
            f"for {command['DocumentName']}. Run aws s3 ls {command['OutputS3KeyPrefix']}/{command_id}/{instance_id}/awsrunShellScript/"
        )

    std = "stderr" if stderr else "stdout"
    key = f"{command['OutputS3KeyPrefix']}/{command_id}/{instance_id}/awsrunShellScript/{doc_path}/{std}"

    s3_client = boto3.client("s3", region_name=config.get("region", None))

    try:
        response = s3_client.get_object(Bucket=bucket, Key=key)
    except ClientError as e:
        if e.response["Error"]["Code"] == "NoSuchKey":
            raise KeyError(f"s3://{bucket}/{key} does not exist")
        else:
            raise e

    # converts body bytes to string lines
    streaming_body = cast(IO[bytes], response["Body"])
    for line in codecs.getreader("utf-8")(streaming_body):
        print(line, end="")

    return None


def fetch_instance_id(config: Config, name: str) -> str:
    if name.startswith("i-"):
        return name

    ec2_client = boto3.client("ec2", region_name=config.get("region", None))
    response = ec2_client.describe_instances(Filters=[{"Name": "tag:Name", "Values": [name]}])

    try:
        return response["Reservations"][0]["Instances"][0]["InstanceId"]
    except IndexError:
        raise ValueError(f"No instance named {name}")


def fetch_instance_ids(config: Config, ids_or_names: List[str]) -> List[str]:

    if ids_or_names == ["all"]:
        return [i["ID"] for i in describe(config)]

    ids: List[str] = []
    names: List[str] = []

    for i in ids_or_names:
        if i.startswith("i-"):
            ids.append(i)
        else:
            names.append(i)

    if names:
        ec2_client = boto3.client("ec2", region_name=config.get("region", None))
        response = ec2_client.describe_instances(Filters=[{"Name": "tag:Name", "Values": names}])

        try:
            for r in response["Reservations"]:
                for i in r["Instances"]:
                    ids.append(i["InstanceId"])
        except IndexError:
            raise ValueError(f"No instances with ids or names {','.join(names)}")
    return ids


def name_filters(names_or_ids: Optional[List[str]] = None) -> List[InstanceInformationStringFilterTypeDef]:
    if names_or_ids and names_or_ids[0].startswith("i-"):
        return [{"Key": "InstanceIds", "Values": names_or_ids}]
    elif names_or_ids:
        return [{"Key": "tag:Name", "Values": names_or_ids}]
    else:
        return []



from __future__ import annotations

import io
from contextlib import redirect_stdout
from typing import Any, Dict, Iterator, List

from moto import mock_ec2
from moto.ec2 import ec2_backends
from moto.ec2.models.amis import AMIS

import aec.command.ec2 as ec2
import aec.util.display as display
from aec.util.config import Config

# fixtures
mock_ec2().start()
region = "ap-southeast-2"
mock_aws_config: Config = {
    "region": region,
    "key_name": "test_key",
    "vpc": {
        "name": "test vpc",
        "subnet": next(ec2_backends[region].get_all_subnets()).id,
        "security_group": "default",
    },
    "iam_instance_profile_arn": "test_profile",
}
ami_id = AMIS[0]["ami_id"]

ec2.launch(mock_aws_config, "alice", ami_id)
ec2.launch(mock_aws_config, "sam", ami_id)


def docs(
    cmd_name: str,
    result: List[Dict[str, Any]] | Iterator[Dict[str, Any]] | Dict | str | None,
) -> str:
    capture = io.StringIO()
    with redirect_stdout(capture):
        display.pretty_print(result)
    return f"{cmd_name}\n{capture.getvalue().rstrip()}"



import os
import shutil

# use importlib_resources backport for compatibility with python < 3.9
import importlib_resources as resources
from importlib_resources.abc import Traversable


def example() -> None:
    """create example config files in ~/.aec/."""

    config_dir = os.path.expanduser("~/.aec/")
    os.makedirs(config_dir, exist_ok=True)

    for r in resources.files("aec.config-example").iterdir():
        copy(r, config_dir)


def copy(res: Traversable, dest_dir: str) -> None:
    if res.name.startswith("__"):
        return
    elif res.is_file():
        with resources.as_file(res) as file:
            dest_file = f"{dest_dir}{res.name}"
            if os.path.exists(dest_file):
                print(f"Skipping {dest_file} (already exists)")
            else:
                print(f"Writing {dest_file}")
                shutil.copy(file, dest_dir)
    elif res.is_dir():
        subdir = f"{dest_dir}{res.name}/"
        os.makedirs(subdir, exist_ok=True)
        for r in res.iterdir():
            copy(r, subdir)
